import numpy as np

# This simple example solves the following constrainted optimization problem:
#
# argmin_{x} \sum_{i = 0}^{4} x_{i}
# s. t. 61 / x_{0}^3
#     + 37 / x_{1}^3
#     + 19 / x_{2}^3
#     + 7 / x_{3}^3
#     + 1 / x_{4}^3
#     - 1 < 0

# The API of the optimizer is entirely defined in the conveks module
import conveks

def get_objective(xFromOpti):
    objective = 0.0624*np.sum(xFromOpti)
    constraints = np.array([61.0/(xFromOpti[0]**3.0)\
        + 37.0/(xFromOpti[1]**3.0)\
        + 19.0/(xFromOpti[2]**3.0)\
        + 7.0/(xFromOpti[3]**3.0)\
        + 1.0/(xFromOpti[4]**3.0)\
        - 1.0])
    return objective,constraints

def get_grad(xFromOpti):
    grad_objective = 0.0624*np.ones(numVariables)
    grad_constraints = np.zeros(numVariables * 1)
    grad_constraints[0] = -3.0*61.0/(xFromOpti[0]**4.0)
    grad_constraints[1] = -3.0*37.0/(xFromOpti[1]**4.0)
    grad_constraints[2] = -3.0*19.0/(xFromOpti[2]**4.0)
    grad_constraints[3] = -3.0*7.0/(xFromOpti[3]**4.0)
    grad_constraints[4] = -3.0*1.0/(xFromOpti[4]**4.0)
    return grad_objective,grad_constraints

# Maximum number of iterations
maxIter = 5

# number of variables and constraints
numVariables = 5

# initial value of the variables in the design space
initialPoint = 5.0*np.ones(numVariables)

# lower bound for design variables (all set to 1)
lowerBound = np.ones(numVariables)

# upper bound for design variables (all set to 10)
upperBound = 10.0 * np.ones(numVariables)

# Initialize conveks
conveks.initialize()

# Initialize the MMA optimizer
conveks.mma.initialize(initialPoint, lowerBound, upperBound)

# Set some general options
conveks.mma.option.setNumber('General.Verbosity', 0)
conveks.mma.option.setNumber('General.SaveHistory', 0)
conveks.mma.option.setNumber('SubProblem.asymptotesRmax', 10)
conveks.mma.option.setNumber('SubProblem.asymptotesRmin', 0.01)
conveks.mma.option.setNumber('SubProblem.adaptSubproblem', 1)
conveks.mma.option.show()

# Get iteration count (here it will be 1 - could be different in case of restart)
it = conveks.mma.getOuterIteration()
innerit = 0;inneritTot = 0; kkt_norm=1.0

# get (copy of) current point
x = conveks.mma.getCurrentPoint()

# compute objective function and constraints
# as well as their sensitivity with respect to design variables at `x'
objective,constraints = get_objective(x)
grad_objective,grad_constraints = get_grad(x)

print('iter.  inner iter.   obj.   max(constr.)   L2-norm(kkt)   point')
print('%3i  %3i  %.6e  %.6e  %.6e'%(it-1, innerit, objective, np.max(np.abs(constraints)), kkt_norm),x)

while it <= maxIter:

    # Solve MMA
    conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints)

    # Evaluate the objective at the new point'
    objective_new,constraints_new = get_objective(conveks.mma.getCurrentPoint())

    # check if the approximation is conservative at the current point
    conserv = conveks.mma.isConservativeSubProblem(objective_new, constraints_new)

    innerit = 0
    while innerit <= 11 and conserv==0:
        innerit += 1

        # adapt the subproblem to make it conservative and generate a new point
        conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints,
            objectiveAtUpdatedPoint=objective_new,constraintsAtUpdatedPoint=constraints_new)

        # evaluate the objective and the constraints at the new point
        objective_new,constraints_new = get_objective(conveks.mma.getCurrentPoint())

        # and check if the subproblem is conservative at this point
        conserv = conveks.mma.isConservativeSubProblem(objective_new, constraints_new)

        # show inner iteration
        print('%3i  %3i '%(it, innerit), '%.6e  %.6e  %.6e'%(objective_new,np.max(np.abs(constraints_new)), kkt_norm),conveks.mma.getCurrentPoint())

    inneritTot += innerit

    # get (copy of) current point
    x = conveks.mma.getCurrentPoint()

    # compute objective function and constraints
    # as well as their sensitivity with respect to design variables at `x'
    objective,constraints = get_objective(x)
    grad_objective,grad_constraints = get_grad(x)

    # evaluate the L2-norm of KKT
    kkt_norm = conveks.mma.getKKTNorm(constraints,grad_objective,grad_constraints)

    # show the current state
    print('%3i  %3i  %.6e  %.6e  %.6e'%(it, innerit, objective, np.max(np.abs(constraints)), kkt_norm),x)

    # get the outer iteration count here
    it = conveks.mma.getOuterIteration()

# show the problem at the converged state
print('*'*80)
print('Optimization has converged after %i iterations.'%(it-1))
print('  > point           :',x)
print('  > objective       : %.4e'%(objective))
print('  > max(constraints): %.4e'%(np.max(np.abs(constraints))))
print('  > L2-norm of kkt  : %.4e'%(kkt_norm))
print('  > inner iterations: %i'%(inneritTot))
print('*'*80)

# This should be called at the end of MMA
conveks.mma.finalize()

# This should be called at the end
conveks.finalize()
