import numpy as np

# This simple example solves the following constrainted optimization problem:
#
# argmin_{x} x^T S x 
# s. t.  n/2 - x^T P x <= 0
#        n/2 - x^T Q x <= 0
#        -1 <= x_j <= 1, j=1,...,n 
#
# where n is the number of design variables, S, P, 
# and Q are symmetric n x n matrices with elements given by
#  s_{ij} = 2 + sin(4*pi*a_{ij}) / (1 + |i-j|)*ln(n),
#  p_{ij} = 1 + 2a_{ij} / (1+|i-j|)ln(n),
#  q_{ij} = 3 - 2a_{ij} / (1+|i-j|)ln(n),
# with a_{ij} = i+j-2 / (2n-2) for all i and j.

# The API of the optimizer is entirely defined in the conveks module
import conveks

# Initialize conveks
conveks.initialize()

# Maximum number of iterations
maxIter = 11

# number of variables and constraints
numVariables = 100

Smat = np.zeros([numVariables,numVariables])
Pmat = np.zeros([numVariables,numVariables])
Qmat = np.zeros([numVariables,numVariables])
lnn = 1.0/np.log(numVariables)
for i in xrange(1,numVariables+1):
    for j in xrange(1,numVariables+1):
        dincr = float((i+j-2))/float((numVariables-1))
        Smat[i-1,j-1] = lnn*(2+np.sin(2*np.pi*dincr))/(1+np.abs(i-j))
        Pmat[i-1,j-1] = lnn*(1+dincr)/(1+np.abs(i-j))
        Qmat[i-1,j-1] = lnn*(3-dincr)/(1+np.abs(i-j))
global Sx,Px,Qx

def get_objective(xFromOpti):
    global Sx,Px,Qx
    Sx = Smat.dot(xFromOpti)
    objective = Sx.dot(xFromOpti)
    Px = Pmat.dot(xFromOpti)
    Qx = Qmat.dot(xFromOpti)
    constraints = np.array([ 0.5*numVariables - Px.dot(xFromOpti), \
      0.5*numVariables - Qx.dot(xFromOpti) ]) 
    return objective,constraints

def get_grad(xFromOpti):
    global Sx,Px,Qx
    grad_objective = 2.0*Sx
    grad_constraints = -2.0*Px
    grad_constraints = np.append(grad_constraints, -2.0*Qx)
    return grad_objective,grad_constraints

# initial value of the variables in the design space
x = 0.5*np.ones(numVariables)

# lower bound for design variables (all set to 1)
lowerBound = -np.ones(numVariables)

# upper bound for design variables (all set to 10)
upperBound = np.ones(numVariables)

# Initialize conveks
conveks.initialize()

# Initialize the MMA optimizer
conveks.mma.initialize(x, lowerBound, upperBound)

# Set some options for MMA
conveks.mma.option.setNumber('General.Verbosity', 4*0)
#conveks.mma.option.setNumber("SubProblem.asymptotesRmax", 10.0)
#conveks.mma.option.setNumber("SubProblem.asymptotesRmin", 0.01)

# Get iteration count (here it will be 1 - could be different in case of restart)
it = conveks.mma.getOuterIteration()
innerit = 0;inneritTot = 0; kkt_norm=1.0

# get (copy of) current point
x = conveks.mma.getCurrentPoint()

# compute objective function and constraints
# as well as their sensitivity with respect to design variables at `x'
objective,constraints = get_objective(x) 
grad_objective,grad_constraints = get_grad(x)
print 'iter.  inner iter.   obj.   max(constr.)   L2-norm(kkt)   point'
print '%3i  %3i   %2.4e   %2.4e   %2.4e'%(it-1, innerit, objective, np.max(np.abs(constraints)), kkt_norm),
print x[0:5]

while it <= maxIter and kkt_norm > 1e-5:

    # Solve MMA
    conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints)

    # Evaluate the objective at the new point'
    objective_new,constraints_new = get_objective(conveks.mma.getCurrentPoint())
    
    # check if the approximation is conservative at the current point
    conserv = conveks.mma.isConservativeSubProblem(objective_new, constraints_new) 

    innerit = 0
    print '\t it.  new-point (xn)       f(xc)       f(xn)      max(c(xc))     max(c(xn))'
    print '\t%3i'%(innerit),
    print conveks.mma.getCurrentPoint()[0:5],
    print '  %.4e  %.4e  %.4e  %.4e'%(objective,objective_new,np.max(constraints), np.max(constraints_new))
    while innerit <= 15 and conserv==0:
        innerit += 1

        # adapt the subproblem to make it conservative and generate a new point
        conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints,
            objectiveAtUpdatedPoint=objective_new,constraintsAtUpdatedPoint=constraints_new)

        # evaluate the objective and the constraints at the new point
        objective_new,constraints_new = get_objective(conveks.mma.getCurrentPoint())

        # show inner iteration
        print '\t%3i'%(innerit),
        print conveks.mma.getCurrentPoint()[0:5],
        print '  %.4e  %.4e  %.4e  %.4e'%(objective,objective_new,np.max(constraints), np.max(constraints_new))

        # and check if the subproblem is conservative at this point
        conserv = conveks.mma.isConservativeSubProblem(objective_new, constraints_new) 
    inneritTot += innerit

    # get (copy of) current point
    x = conveks.mma.getCurrentPoint()

    # compute objective function and constraints
    # as well as their sensitivity with respect to design variables at `x'
    objective,constraints = get_objective(x) 
    grad_objective,grad_constraints = get_grad(x)

    # evaluate the L2-norm of KKT 
    kkt_norm = conveks.mma.getKKTNorm(constraints,grad_objective,grad_constraints)

    # show the current state
    print '%3i  %3i  %2.4e  %2.4e  %2.4e'%(it, innerit, objective, np.max(np.abs(constraints)), kkt_norm),
    print x[0:5]

    # get the outer iteration count here
    it = conveks.mma.getOuterIteration()

# show the problem at the converged state 
print '*'*80
print 'Optimization has converged after %i iterations.'%(it-1)
print '  > point           :',x[0:5]
print '  > objective       : %.4e'%(objective)
print '  > max(constraints): %.4e'%(np.max(np.abs(constraints)))
print '  > L2-norm of kkt  : %.4e'%(kkt_norm)
print '  > inner iterations: %i'%(inneritTot)
print '*'*80

# This should be called at the end of MMA
conveks.mma.finalize()

# This should be called at the end 
conveks.finalize()