import numpy as np

# This simple example solves the following constrainted optimization problem:
#
# argmin_{x} x^T S x
# s. t.  n/2 - x^T P x <= 0
#        n/2 - x^T Q x <= 0
#        -1 <= x_j <= 1, j=1,...,n
#
# where n is the number of design variables, S, P,
# and Q are symmetric n x n matrices with elements given by
#  s_{ij} = 2 + sin(4*pi*a_{ij}) / (1 + |i-j|)*ln(n),
#  p_{ij} = 1 + 2a_{ij} / (1+|i-j|)ln(n),
#  q_{ij} = 3 - 2a_{ij} / (1+|i-j|)ln(n),
# with a_{ij} = i+j-2 / (2n-2) for all i and j.

# The API of the optimizer is entirely defined in the conveks module
import conveks

# Maximum number of iterations
maxIter = 11

# number of variables and constraints
numVariables = 100

Smat = np.zeros([numVariables,numVariables])
Pmat = np.zeros([numVariables,numVariables])
Qmat = np.zeros([numVariables,numVariables])
lnn = 1.0/np.log(numVariables)
for i in range(1,numVariables+1):
    for j in range(1,numVariables+1):
        dincr = float((i+j-2))/float((numVariables-1))
        Smat[i-1,j-1] = lnn*(2+np.sin(2*np.pi*dincr))/(1+np.abs(i-j))
        Pmat[i-1,j-1] = lnn*(1+dincr)/(1+np.abs(i-j))
        Qmat[i-1,j-1] = lnn*(3-dincr)/(1+np.abs(i-j))
global Sx,Px,Qx

def get_objective(xFromOpti):
    global Sx,Px,Qx
    Sx = Smat.dot(xFromOpti)
    objective = -Sx.dot(xFromOpti)
    Px = Pmat.dot(xFromOpti)
    Qx = Qmat.dot(xFromOpti)
    constraints = np.array([ Px.dot(xFromOpti) - 0.5*numVariables, \
      Qx.dot(xFromOpti) - 0.5*numVariables ])
    return objective,constraints

def get_grad(xFromOpti):
    global Sx,Px,Qx
    grad_objective = -2.0*Sx
    grad_constraints = 2.0*Px
    grad_constraints = np.append(grad_constraints, 2.0*Qx)
    return grad_objective,grad_constraints

# initial value of the variables in the design space
x = 0.25*np.ones(numVariables)

# lower bound for design variables (all set to 1)
lowerBound = -np.ones(numVariables)

# upper bound for design variables (all set to 10)
upperBound = np.ones(numVariables)

# Initialize conveks
conveks.initialize()

# Initialize the MMA optimizer
conveks.mma.initialize(x, lowerBound, upperBound)

# Set some options for MMA
conveks.mma.option.setNumber('General.Verbosity', 0)
#conveks.mma.option.setNumber("SubProblem.asymptotesRmax", 10.0)
#conveks.mma.option.setNumber("SubProblem.asymptotesRmin", 0.01)

# Get iteration count (here it will be 1 - could be different in case of restart)
it = conveks.mma.getOuterIteration()

# L2-norm of the KKT
kkt_norm = 1.0

# get (copy of) current point
x = conveks.mma.getCurrentPoint()

# compute objective function and constraints
# as well as their sensitivity with respect to design variables at `x'
objective,constraints = get_objective(x)
grad_objective,grad_constraints = get_grad(x)

print('iter.   obj.   max(constr.)   L2-norm(kkt)   point')
print('%3i  %2.4e  %2.4e  %2.4e'%(it-1, objective, np.max(np.abs(constraints)), kkt_norm),x[0:5])

while it <= maxIter:

    # call MMA and update the current point
    conveks.mma.updateCurrentPoint(objective,grad_objective,constraints,grad_constraints)

    # get the count of outer iterations
    it = conveks.mma.getOuterIteration()

    # get (copy of) current point
    x = conveks.mma.getCurrentPoint()

    # compute objective function and constraints
    # as well as their sensitivity with respect to design variables at `x'
    objective,constraints = get_objective(x)
    grad_objective,grad_constraints = get_grad(x)

    # evaluate the L2-norm of KKT
    kkt_norm = conveks.mma.getKKTNorm(constraints,grad_objective,grad_constraints)

    print('%3i  %2.4e  %2.4e  %2.4e'%(it-1, objective, np.max(np.abs(constraints)), kkt_norm),x[0:5])

# evalute the problem at the optimal point
xc = conveks.mma.getCurrentPoint()
obj_c,constr_c = get_objective(xc)
gobj_c,gconstr_c = get_grad(xc)
kkt_norm = conveks.mma.getKKTNorm(constr_c,gobj_c,gconstr_c)
print('*'*80)
print('Optimization has converged after %i iterations.'%(it-1))
print('  > point         : ',xc[0:5])
print('  > objective     : ',obj_c)
print('  > constraints   : ',constr_c)
print('  > L2-norm of kkt: ',kkt_norm)
print('*'*80)

# This should be called at the end of MMA
conveks.mma.finalize()

# This should be called at the end
conveks.finalize()
